/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"

struct application application;

struct ble_api_srv_profile *server;
struct ble_api_cli_profile *client;

struct nvm_partition ble_partitions[] = {
    {nvm_reserved, 4096},            /* A-side header */
    {nvm_application_a0, 7 * 4096},  /* A0 application*/
    {nvm_application_a1, 12 * 4096}, /* A1 application*/
    {nvm_reserved, 4096},            /* B-side header */
    {nvm_application_b0, 7 * 4096},  /* B0 application*/
    {nvm_application_b1, 12 * 4096}, /* B1 application*/
    {nvm_key_store_data_1, 4096},    /* Key store data 1 */
    {nvm_key_store_data_2, 4096},
    /* Key store data 2 */
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

extern struct nvm_partition other_partitions[];
extern struct nvm_partition system_partitions[];

struct nvm_partition *const nvm_partitions[] = {
    &system_partitions[0], &ble_partitions[0], &other_partitions[0], NULL, /* end marker - don't remove */
};

#include "bt_api_ven_flash.h"

int board_nvm_open_flash(uint32_t write_protection)
{
  return bt_api_ven_nvm_storage_open(write_protection);
}

int board_nvm_close_flash(void)
{
  return bt_api_ven_nvm_storage_close();
}

int board_nvm_read_flash(uint32_t address, uint8_t *data, uint32_t length)
{
  return bt_api_ven_nvm_storage_read(address, data, (uint16_t)length);
}

int board_nvm_sector_erase_flash(uint32_t address)
{
  return bt_api_ven_nvm_storage_sector_erase(address);
}

int board_nvm_write_flash(uint32_t address, uint8_t *data, uint32_t length)
{
  return bt_api_ven_nvm_storage_write(address, data, (uint16_t)length);
}

/* This is a simple ringbuffer implementation. It is mainly used to store any
"keypresses" that are received from the UART for the later forwarding to an
existing Bluetooth connection. */
void flush_rb(struct ringbuffer *rb)
{
  memset(rb, 0, sizeof(*rb));
}

int add_to_rb(struct ringbuffer *rb, uint8_t c)
{
  ring_pos_t next_head = (rb->ring_head + 1) % RING_SIZE;
  if (next_head != rb->ring_tail) {
    /* there is room */
    rb->ring_data[rb->ring_head] = c;
    rb->ring_head = next_head;
    return 0;
  }

  /* no room left in the buffer */
  return -1;
}

int remove_from_rb(struct ringbuffer *rb, uint8_t *c)
{
  ring_pos_t ring_head = rb->ring_head;

  if (ring_head != rb->ring_tail) {
    *c = rb->ring_data[rb->ring_tail];
    rb->ring_tail = (rb->ring_tail + 1) % RING_SIZE;
    return 0;
  }

  return -1;
}

int peek_into_rb(struct ringbuffer *rb, uint8_t *c)
{
  ring_pos_t ring_head = rb->ring_head;

  if (ring_head == rb->ring_tail) {
    return -1;
  }

  *c = rb->ring_data[rb->ring_tail];
  return 0;
}

int application_handle_keyboard_input(int key)
{
  add_to_rb(&application.in, (uint8_t)key);
  return 0;
}

#ifdef SPEED_TEST
static int speed_test_callback(const void *argument)
{
  static int last_tx = 0;
  static int last_rx = 0;
  int ret;

  if (application.is_configuration_enabled == false) {
    application.timer_speed_test.function = NULL;
    return 0;
  }

  if (application.rx.is_enabled == false && application.tx.is_enabled == false) {
    application.timer_speed_test.function = NULL;
    return 0;
  }

  xprintf("rx %d (%d), tx %d (%d)\n", (application.rx.total_bytes - last_rx), application.rx.total_bytes,
          (application.tx.total_bytes - last_tx), application.tx.total_bytes);

  last_tx = application.tx.total_bytes;
  last_rx = application.rx.total_bytes;

  ret = timer_start(&application.timer_speed_test, 1000);
  if (ret) {
    warning("timer_start() failed\n");
    return -1;
  }

  return 0;
}

static int speed_test_tx(void)
{
  int len;
  int ret;

  if (!application.is_spp_connected) {
    //		dprintf("not connected - doing nothing\n");
    return 0;
  }

  if (application.tx.is_enabled == false) {
    return 0;
  }

  if (application.is_sending == true) {
    return 0;
  }

  len = application.mtu_size;

  ret = configuration_speed_test_tx(application.buf, len, application.connection.connection_handle);
  if (ret) {
    dprintf("configuration_speed_test_tx() failed\n");
    return 0;
  }

  application.tx.total_bytes += len;
  return 0;
}

int application_process(void)
{
  uint8_t c;
  int ret;

  if (application.tx.is_enabled == true) {
    ret = speed_test_tx();
    if (ret) {
      warning("speed_test_tx() failed\n");
      /* fall throgh */
    }
  }

  dprintf("\n");

  ret = remove_from_rb(&application.in, &c);
  if (ret) {
    return 0;
  }

  if (c == 't') {
    application.tx.is_enabled = !application.tx.is_enabled;
  } else if (c == 'r') {
    application.rx.is_enabled = !application.rx.is_enabled;
  } else if (c == 's') {
    /* start application specific handling */
    if (application.is_configuration_enabled == false) {
      ret = configuration_start();
    } else {
      ret = configuration_stop();
    }
    application.is_configuration_enabled = !application.is_configuration_enabled;
  } else if (c == 'd') {
    /* disconnct, if connected */
  }

  if (application.is_configuration_enabled && (application.rx.is_enabled || application.tx.is_enabled)) {
    if (application.timer_speed_test.function == NULL) {
      application.timer_speed_test.function = speed_test_callback;
      ret = timer_start(&application.timer_speed_test, 1000);
      if (ret) {
        warning("timer_start() failed\n");
        /* fall through */
      }
    }
  }

  xprintf("t %d, r %d, c %d\n", application.tx.is_enabled, application.rx.is_enabled, application.is_configuration_enabled);

  return 0;
}
#endif
