/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"
#include "ble_api_master.h"

#include "client_spp_over_ble_api.h"

#ifdef SPEED_TEST
#define application_process configuration_process
#endif

/*! Structure that contains central-specific application data */
struct config {
  bool is_scan_stop_requested;         /*!< Indicates if scanning should be stopped */
  bool is_scanning;                    /*!< Indicates if the device is currently scanning */
  bool is_name_found;                  /*!< Indicates if the correct name has been found */
  bool is_uuid_found;                  /*!< Indicates if the correct UUID has been found */
  bool is_disconnect_requested;        /*!< Indicates if disconnect is already in progress */
  uint8_t bd_addr[6];                  /*!< Bluetooth device address of device to connect to */
  enum tcu_smp_address_type addr_type; /*!< Address type of device to connect to */
};

static struct config config;

static uint8_t *name;
static uint16_t name_len;

/* static function forward declarations */
static int scan_start(void);
static int scan_stop(void);

int application_init(const uint8_t *device_name, int device_name_len)
{
  enum ble_api_result result;

  client = &spp_over_ble_cli_profile;

  name_len = device_name_len;
  name = (uint8_t *)device_name;

  static struct ble_api_legacy_connection_parameters legacy_connection_parameters = {
    0x06, // LE_CONNECTION_INTERVAL_MINIMUM,
    0x06, // LE_CONNECTION_INTERVAL_MAXIMUM,
    0x00, // LE_CONNECTION_LATENCY,
    100,  // LE_SUPERVISION_TIMEOUT,
    LE_MINIMUM_CE_LENGTH,
    LE_MAXIMUM_CE_LENGTH,
    LE_INITIATOR_FILTER_POLICY
  };

  result = ble_api_configure_legacy_connection_parameters(&legacy_connection_parameters);
  if (result != LE_API_SUCCESS) {
    return -1;
  }

  return 0;
}

int configuration_start(void)
{
  int ret;

  ret = scan_start();
  if (ret) {
    dprintf("scan_start() failed\n");
    return -1;
  }

  xprintf("CENTRAL\n");

  return 0;
}

int application_process(void)
{
  static int once = 0;
  int len = 0;
  //  enum ble_api_result result;
  int ret;

  if (!once) {
    once = 1;

    ret = configuration_start();
    if (ret) {
      dprintf("configuration_start() failed\n");
      return -1;
    }
  }

#if 0
  /* check if encryption is setup correctly */
  if (application.is_connected == true) {
    if (application.encryption_setup_failed == true) {
      if (config.is_disconnect_requested == false) {
        result = ble_api_disconnect(&application.connection.bd_addr[0]);
        if (result != LE_API_SUCCESS) {
          dprintf("ble_api_disconnect() failed\n");
          return -1;
        }
        config.is_disconnect_requested = true;
      }
    }
  }
#endif

  if (application.is_spp_connected == false) {
    //    dprintf("spp not connected - doing nothing\n");
    return 0;
  }

#if 0
  if (application.is_encrypted == false) {
//    dprintf("not encrypted - doing nothing\n");
    return 0;
  }
#endif

  if (application.is_sending == true) {
    dprintf("currently sending - doing nothing\n");
    return 0;
  }

  while (len < application.mtu_size) {
    ret = remove_from_rb(&application.in, &application.buf[len]);
    if (ret) {
      break;
    }
    len++;
  }

  if (!len) {
    //    dprintf("nothing to send - doing nothing\n");
    return 0;
  }

  ret = client_spp_over_ble_service_write_start(application.buf, len, application.connection.connection_handle);
  if (ret) {
    dprintf("client_spp_over_ble_service_write_start() failed\n");
    return -1;
  }

  application.is_sending = true;

  //	dprintf("sending data\n");

  return 0;
}

static int scan_start(void)
{
  enum ble_api_result result;

  struct ble_api_uuid uuid = {client_service_spp_over_ble.rom.uuid_type};
  memcpy(&uuid.uuid, client_service_spp_over_ble.rom.uuid, LE_LONG_UUID);

  static struct ble_api_legacy_scan_parameters legacy_scan_parameters = {
    0x20, // LE_SCAN_INTERVAL,
    0x20, // LE_SCAN_WINDOW,
    LE_SCAN_FILTER_POLICY,
    0x00, // LE_SCAN_DUPLICATION_FILTER
    /* scan type is set explicitely below */
  };


  /* if no name is given, we will let the SDK filter for the SPPoverBLE UUID and
  connect to the first device that is found. */
  if (!name) {
    legacy_scan_parameters.scan_type = LE_PASSIVE;

    result = ble_api_configure_legacy_scan_parameters(&legacy_scan_parameters);
    if (result != LE_API_SUCCESS) {
      return -1;
    }

    result = ble_api_master_start_scan_filter_uuid(&uuid, ADDRESS_CONFIGURATION_PUBLIC);
    if (result) {
      dprintf("ble_api_master_start_scan_filter_none() failed\n");
      return -1;
    }
  }
  /* otherwise we will just filter for all devices and do the filtering for both
  the UUID and the name manually later. this is necessary, because UUID and
  device name may arrive in the advertising data or the scan response
  separately... */
  else {
    legacy_scan_parameters.scan_type = LE_ACTIVE;

    result = ble_api_configure_legacy_scan_parameters(&legacy_scan_parameters);
    if (result != LE_API_SUCCESS) {
      return -1;
    }

    result = ble_api_master_start_scan_filter_none(ADDRESS_CONFIGURATION_PUBLIC);
    if (result) {
      dprintf("ble_api_master_start_scan_filter_none() failed\n");
      return -1;
    }
  }

  dprintf("\n");
  return 0;
}

static int scan_stop(void)
{
  enum ble_api_result result;

  result = ble_api_master_scan_disable();
  if (result) {
    dprintf("ble_api_master_scan_disable() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

static enum ble_api_result scan_enable_response_callback(uint8_t status)
{
  dprintf("status %02x\n", status);

  config.is_scanning = true;

  return LE_API_SUCCESS;
}

static enum ble_api_result scan_disable_response_callback(uint8_t status)
{
  enum ble_api_result result;

  dprintf("status %02x\n", status);
  config.is_scanning = false;

  if (config.is_scan_stop_requested == true) {
    dprintf("is_scan_stop_requested %d, doing nothing", config.is_scan_stop_requested);
    return LE_API_SUCCESS;
  }

  dprintf("connecting to " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_LE(config.bd_addr));

  result = ble_api_master_connect(config.bd_addr, config.addr_type, SMP_ADDRESS_TYPE_PUBLIC_ADDRESS, 5);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_master_connect() failed\n");
    return LE_API_ERROR;
  }

  return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_scan_status_update_callback(enum ble_api_scan_operation operation, uint8_t status)
{
  if (operation == SCAN_START) {
    return scan_enable_response_callback(status);
  } else if (operation == SCAN_STOP) {
    return scan_disable_response_callback(status);
  }

  /* unexpected ble_api_scan_operation... */
  return LE_API_ERROR;
}

enum ble_api_result ble_application_master_advertise_report_event_callback(struct tcu_mng_le_adv_report_event *event)
{
  struct ble_api_extract_adv_data extract_adv_data_type;
  enum ble_api_result result;
  int ret;

  /* it may happen that we already requested a scan stop, but advertise reports will come in. ignore them */
  if (config.is_name_found == true && config.is_uuid_found == true) {
    return LE_API_SUCCESS;
  }

  /* if we have been given a (device) name, this is the name we need to check
  the found device against. only if it matches we will connect to that device.
*/
  if (name) {

    result = ble_api_extract_uuid_data(LE_LONG_UUID, event->adv_data, event->adv_data_len, &extract_adv_data_type);
    if (result == LE_API_SUCCESS) {
      //      xprintf("UUID @ " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_BE(event->address));
      ret = memcmp(client_service_spp_over_ble.rom.uuid, extract_adv_data_type.data, LE_LONG_UUID);
      if (!ret) {
        config.is_uuid_found = true;
      }
    }

    result = ble_api_extract_device_name(event->adv_data, event->adv_data_len, &extract_adv_data_type);
    if (result == LE_API_SUCCESS) {
      //      xprintf("NAME '%.*s' @ " BD_ADDR_FORMAT_STRING "\n", extract_adv_data_type.data_len, extract_adv_data_type.data,
      //      BD_ADDR_ARGUMENTS_BE(event->address));
      if (extract_adv_data_type.data_len == name_len) {
        ret = memcmp(name, extract_adv_data_type.data, name_len);
        if (!ret) {
          config.is_name_found = true;
        }
      }
    }

#pragma message("Warning: This is too naive. The Bluetooth address must be stored and taken into consideration as well.")
    if (config.is_name_found == false || config.is_uuid_found == false) {
      return LE_API_SUCCESS;
    }
  }

  /* store peer information for later connect */
  memcpy(&config.bd_addr[0], event->address, 6);
  config.addr_type = event->address_type;

  ret = scan_stop();
  if (ret) {
    dprintf("scan_stop() failed\n");
    return LE_API_ERROR;
  }

  dprintf("\n");
  return LE_API_SUCCESS;
}

void client_descriptor_spp_over_ble_enable_disable_notification_indication_status_callback(uint8_t operation, uint8_t status,
                                                                                           uint16_t connection_handle)
{
  /* fixme: check operation and status here */

  application.is_spp_connected = true;
  xprintf("SPP %d\n", application.mtu_size);

  dprintf("connection_handle %04x\n", connection_handle);
}

void client_descriptor_spp_over_ble_write_done_callback(uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
}

void client_spp_over_ble_indication_callback(uint8_t *value, uint16_t count, uint16_t connection_handle)
{
#ifdef SPEED_TEST
  if (application.rx.is_enabled == true) {
    application.rx.total_bytes += count;
    return;
  }
#endif
  xprintf("%.*s", count, value);
}

void client_spp_over_ble_write_done_callback(uint16_t connection_handle)
{
  application.is_sending = false;

  if (application.is_spp_connected) {
    return;
  }

  dprintf("connection_handle %04x\n", connection_handle);
}

void client_spp_over_ble_write_failed_callback(uint8_t status, uint16_t connection_handle)
{
  application.is_sending = false;
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);
}

void client_spp_over_ble_disconnected_callback(uint16_t connection_handle)
{
  int ret;

  dprintf("connection_handle %04x\n", connection_handle);

  flush_rb(&application.in);

  config.is_uuid_found = false;
  config.is_name_found = false;
  config.is_disconnect_requested = false;
  application.is_spp_connected = false;
  application.is_sending = false;

  ret = scan_start();
  if (ret) {
    dprintf("scan_start() failed\n");
    return;
  }

  xprintf("\nDISCONNECT\n");
}

void ble_application_mng_set_data_length_response_callback(uint8_t status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);
}

void client_spp_over_ble_profile_found_callback(uint16_t connection_handle, uint16_t mtu_size)
{
  int ret;

  application.mtu_size = mtu_size;

  ret = client_spp_over_ble_service_enable_indication(connection_handle);
  if (ret) {
    dprintf("client_spp_over_ble_service_enable_indication() failed\n");
    return;
  }

  dprintf("connection_handle %04x, mtu_size %d\n", connection_handle, mtu_size);
  xprintf("MTU %d\n", application.mtu_size);
}

#ifdef SPEED_TEST
int configuration_stop(void)
{
  int ret;

  config.is_scan_stop_requested = true;

  ret = scan_stop();
  if (ret) {
    dprintf("scan_stop() failed\n");
    return -1;
  }

  xprintf("STOP\n");

  return 0;
}

int configuration_speed_test_tx(uint8_t *data, uint16_t length, uint16_t connection_handle)
{
  int ret;

  ret = client_spp_over_ble_service_write_start(data, length, connection_handle);
  if (ret) {
    dprintf("client_spp_over_ble_service_write_start() failed\n");
    return -1;
  }

  application.is_sending = true;
  return 0;
}
#endif

void ble_application_slave_set_advertise_data_response_callback(uint8_t status)
{
  dprintf("\n");
  __debugbreak();
}

void ble_application_slave_get_advertise_data_callback(uint8_t **adv_data_ptr, uint16_t *adv_data_length)
{
  dprintf("\n");
  __debugbreak();
}

void ble_application_slave_get_scan_response_data_callback(uint8_t **adv_data_ptr, uint16_t *adv_data_length)
{
  dprintf("\n");
  __debugbreak();
}
