/****************************************************************************
*   COPYRIGHT 2017 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

/* fixme: add data length extension */

#include "application.h"
#include "ble_api.h"
#include "bt_api_ven_flash.h"
#include "debug.h"
#include "host_api.h"
#include "lp_function.h"
#include "osal.h"
#include "patch_api.h"
#include "system.h"
#include "uart1_function.h"

/* This does not belong here */
#define NVM_SIZE 262144
uint8_t init_done = 0;

static uint8_t device_name[] = "PAN1760A-USB";
static int device_name_len = sizeof(device_name) - 1;

/* the following is TC35678 specific and probably should not be here */
#include "internal_apis.h"

struct nvm_partition system_partitions[] = {
    {nvm_reserved, 2},
    {nvm_bd_address, 6},
    {nvm_reserved, 4088},
    {nvm_bt_sw_stack, 61440},
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition other_partitions[] = {
    {nvm_user_area, 13 * 4096},
    {nvm_reserved, 4096},
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

/* This function reads the pre-programmed Bluetooth device address from the
built-in flash. Surprisingly, this also initializes the Bluetooth core with
this address. So this procedure is mandatory to be executed. */
static int init_bd_address(void)
{
#define BD_ADDR_LEN 0x06
#define BD_ADDR_OFFSET 0x02
  uint8_t bd_address[BD_ADDR_LEN];
  int ret;

  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageReadSpecifiedAddr(BD_ADDR_LEN, BD_ADDR_OFFSET);
  if (ret) {
    dprintf("SYS_API_QSPIStorageReadSpecifiedAddr() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageRead(BD_ADDR_OFFSET, BD_ADDR_LEN, &bd_address[0]);
  if (ret) {
    dprintf("SYS_API_QSPIStorageRead() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

static void handle_low_power_event(void)
{
  uint16_t pmu_cause;
  uint16_t crw_rst_cause;
  int ret;

  ret = lp_get_wake_up_cause(&pmu_cause, &crw_rst_cause);
  if (ret) {
    warning("lp_get_wake_up_cause() failed\r\n");
    return;
  }

  if (pmu_cause & ((1 << 2) | (1 << 5))) {
    warning("Wake up from Backup mode\n");
  } else {
    warning("Wake up from Sleep mode\n");
  }
  if (pmu_cause & (1 << 6)) {
    warning("Wake up because advertized specified times\n");
  }
  if (pmu_cause & (1 << 7)) {
    warning("Wake up because of BLE connection request\n");
  }
  if (pmu_cause & (1 << 8)) {
    warning("Wake up because of GPIO0 level change\n");
  }
  if (pmu_cause & (1 << 9)) {
    warning("Wake up because of GPIO15 level change\n");
  }
  if (pmu_cause & (1 << 11)) {
    warning("Wake up because of ANYKEY function\n");
  }
  if ((pmu_cause & (1 << 12)) && (pmu_cause & (1 << 2))) {
    warning("Wake up because 10 hours passed\n");
  }
  if (pmu_cause & (1 << 13)) {
    warning("Wake up because of RTC\n");
  }
}

#ifdef __GNUC__
void __iar_data_init3(void)
{
  extern char _bstart;
  extern char _bend;
  char *dst;

  /* Zero bss */
  for (dst = &_bstart; dst < &_bend; dst++) {
    *dst = 0;
  }
}

__attribute__((section("ENTRY_FUNCTION"))) __attribute__((noreturn))
#endif
#if defined(__IAR_SYSTEMS_ICC__)
extern void
__iar_data_init3(void);
#pragma default_function_attributes = @"ENTRY"
#endif
/* This is the main function. It does the basic initialization and then goes
into the main message processing loop. The "system" and "message" handling
functions come straight from the Bluetooth SDK. */
void main(void)
{
  enum tsb_bluetooth_chip_version version;
  enum ble_api_result result;
  unsigned int flag = 0;
  int ret;

  __iar_data_init3();

  initializeUart1();

  ret = init_bd_address();
  if (ret) {
    dprintf("init_bd_address() failed\n");
    goto fatal_error;
  }

  ret = system_read_version(&version);
  if (ret) {
    dprintf("system_read_version() failed\n");
    goto fatal_error;
  }

  patch_api_update_bt_firmware(version);

  result = ble_api_init(version);
  if (result) {
    dprintf("ble_api_init() failed\n");
    goto fatal_error;
  }

#define BLE_INIT 0x01
  init_done |= BLE_INIT;

  ret = nvm_init(NVM_SIZE);
  if (ret) {
    dprintf("nvm_init() failed\n");
    goto fatal_error;
  }

  ret = key_store_ble_init();
  if (ret) {
    dprintf("key_store_ble_init() failed\n");
    goto fatal_error;
  }

  /* The peripheral and central demo need different init functions */
  ret = application_init(device_name, device_name_len);
  if (ret) {
    dprintf("application_init() failed\n");
    goto fatal_error;
  }

  while (1) {

    application_process();

    event_wait(&flag);

    if (flag & (C2_GPIO_WAKEUP_BIT | C2_GPIO_WAKEUP_2_BIT | C2_AUTO_WAKEUP_BIT)) {
      handle_low_power_event();
    }

    if ((flag & C2_RTC_WAKEUP_BIT)) {
    }

    if ((flag & C2_MSG_INT_BIT)) {
      /* ble_api_message_handler is called if received event is BLE event */
      ret = ble_api_message_handler(server, client);
      if (ret) {
        dprintf("ble_api_message_handler() failed\n");
        /* fall through */
      }
    }

    if ((flag & USER_INTERRUPT_BIT)) {
      /* application processing */
      warning("GPIO Interrupt received\n");
    }
  }

fatal_error:
  warning("fatal error - application did not start\n");
  while (1) {
    /* fixme */
  }
}
#if defined(__IAR_SYSTEMS_ICC__)
#pragma default_function_attributes =
#endif

void ble_application_event_handling_failed_callback(uint8_t service_id, uint8_t opcode, uint8_t error)
{
  warning("service_id 0x%02x, opcode 0x%02x, error 0x%02x\n", service_id, opcode, error);
}

#include "mini-printf.h"

static uint8_t buf[128];
void uart1_printf(const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = mini_vsnprintf((char *)buf, sizeof(buf), format, args);
  va_end(args);

  sendUart1(buf, ret);
}

void tcu_message_debug_output(const char *prefix, const uint8_t *buf, uint32_t length)
{
#if 0
  static char debug_buf[256];
  
  int i;
  char *ptr = &debug_buf[0];
  int len = sizeof(debug_buf);
  int ret;
  
  ret = mini_snprintf(ptr, len, "%03d %s ", length, prefix);
  ptr += ret; len -= ret;
  
  for (i = 0; i < length; i++) {
    ret = mini_snprintf(ptr, len, "%02x ", buf[i]);
    ptr += ret;  len -= ret;
    if (!len) {
      break;
    }
  }

  uart1_printf("%s\n", debug_buf);
#endif
}

void data_receive_debug_callback(uint8_t *buf, uint32_t length)
{
  tcu_message_debug_output("<--", buf, length);
}

void ble_application_default_event_callback(struct cmn_msg *cmn_msg, uint8_t *data)
{
  warning("\n");
}

void ble_application_uart_message_callback(struct cmn_msg *cmn_msg, uint8_t *data)
{
  application_process();
}

extern int message_timer_callback(void *timer);

void ble_application_timer_message_callback(struct cmn_msg *cmn_msg, uint8_t *data)
{
  message_timer_callback((void *)cmn_msg->user_data);
}

int uart1_receive_bytes_callback_isr(uint8_t *ptr, int len)
{
  int i;
  int ret;

  for (i = 0; i < len; i++) {
    ret = application_handle_keyboard_input(ptr[i]);
    if (ret) {
      return -1;
    }
  }

  return 0;
}
