/*
 * COPYRIGHT (C) 2018
 * TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION ALL RIGHTS RESERVED
 *
 * THE SOURCE CODE AND ITS RELATED DOCUMENTATION IS PROVIDED "AS IS". TOSHIBA
 * ELECTRONIC DEVICES & STORAGE CORPORATION MAKES NO OTHER WARRANTY OF ANY KIND,
 * WHETHER EXPRESS, IMPLIED OR, STATUTORY AND DISCLAIMS ANY AND ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY, SATISFACTORY QUALITY, NON INFRINGEMENT AND
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * THE SOURCE CODE AND DOCUMENTATION MAY INCLUDE ERRORS. TOSHIBA ELECTRONIC
 * DEVICES & STORAGE CORPORATION RESERVES THE RIGHT TO INCORPORATE MODIFICATIONS
 * TO THE SOURCE CODE IN LATER REVISIONS OF IT, AND TO MAKE IMPROVEMENTS OR
 * CHANGES IN THE DOCUMENTATION OR THE PRODUCTS OR TECHNOLOGIES DESCRIBED
 * THEREIN AT ANY TIME.
 *
 * TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION SHALL NOT BE LIABLE FOR ANY
 * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGE OR LIABILITY ARISING FROM YOUR USE
 * OF THE SOURCE CODE OR ANY DOCUMENTATION, INCLUDING BUT NOT LIMITED TO, LOST
 * REVENUES, DATA OR PROFITS, DAMAGES OF ANY SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL NATURE, PUNITIVE DAMAGES, LOSS OF PROPERTY OR LOSS OF PROFITS
 * ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT, OR BEING UNUSABLE, EVEN
 * IF ADVISED OF THE POSSIBILITY OR PROBABILITY OF SUCH DAMAGES AND WHETHER
 * A CLAIM FOR SUCH DAMAGE IS BASED UPON WARRANTY, CONTRACT, TORT, NEGLIGENCE OR
 * OTHERWISE.
 */

#include "hci_api.h"
#include "hci_api_internal.h"
#include "uart_api.h"
#include "debug_api.h"
#include "versions.h"

int hci_mode_active = 1;

/*! HCI event packet type */
#define HCI_PACKET_TYPE_EVENT_PACKET 0x04

/*! Helper macro to calculate the length of a HCI response */
#define HCI_API_RETRIEVE_HCI_RESPONSE_LENGTH(ptr) (ptr[2] + 3)

/*! Helper macro to ccheck the correct type of a HCI response */
#define HCI_API_CHECK_HCI_RESPONSE(ptr) (ptr[0] != HCI_PACKET_TYPE_EVENT_PACKET)

#define SET_OCF_OGF(x) \
  (x >> 8) & 0xff, (x & 0xff)

#define OCF_OGF_HCI_RESET 0x030c
#define OCF_OGF_HCI_WRITE_BD_ADDRESS 0x1310

#define HCI_EVENT_CODE_COMMAND_COMPLETE_EVENT 0x0e

#define CMD_I2C_ENABLE 0x5b
#define CMD_EEPROM_WRITE_ENABLE 0x83
#define CMD_EEPROM_READ 0x88
#define CMD_EEPROM_WRITE 0x89
#define CMD_GET_FIRMWARE_VERSION 0x0d
#define CMD_SET_DEEP_SLEEP 0x68
#define CMD_BLE_SET_RTSCTS 0xa2
#define CMD_BLE_SET_TRANSMIT_POWER 0x9b
#define CMD_FLASH_OPEN_STORAGE 0xb3
#define CMD_FLASH_READ_WRITE_STORAGE 0xb4
#define CMD_FLASH_ERASE_SECTOR_IN_STORAGE 0xb7
#define CMD_FLASH_ERASE_STORAGE_DEVICE 0x06
#define CMD_FLASH_CLOSE_STORAGE 0xb9

#define HCI_RTS_CTS 0x93
#define HCI_SET_MODE 0x99
#define HCI_LOC_DBUS_WRITE_CMD 0xc2

#define M2_SET_CMD 0xa0
#define M2_GET_CMD 0xa1

static int check_m2(uint8_t cmd);

struct hci_command_complete_event
{
  uint8_t packet_type;
  uint8_t event_code;
  uint8_t command_length;
  uint8_t num_hci_packets;
  uint16_t ocf_ogf;
  uint8_t status;
};

struct hci_command_status_event
{
  uint8_t packet_type;
  uint8_t event_code;
  uint8_t command_length;
  uint8_t status;
  uint8_t num_hci_packets;
  uint8_t ocf;
  uint8_t ocf_ogf;
};

struct hci_command_event
{
  uint8_t packet_type;
  uint8_t event_code;
  uint8_t command_length;
  uint8_t reserved1;
  uint8_t reserved2;
  uint8_t command;
  uint8_t status;
};

struct hci_m2_xet_event
{
  uint8_t packet_type;
  uint8_t event_code;
  uint8_t command_length;
  uint8_t reserved1;
  uint8_t reserved2;
  uint8_t command;
  uint8_t reserved[3];
  uint8_t acceptor;
  uint8_t information_id;
  uint8_t command_result;
  uint8_t data_type;
  uint8_t *data;
};

static int parse_hci_command_complete_event(uint8_t *buf, int len, struct hci_command_complete_event *event)
{
  if (len != 7) {
    hci_bs_printf("len != 7\n");
    return -1;
  }

  event->packet_type = *buf++;
  if (event->packet_type != HCI_PACKET_TYPE_EVENT_PACKET) {
    hci_bs_printf("HCI_PACKET_TYPE_EVENT_PACKET failed\n");
    return -1;
  }
  event->event_code = *buf++;
  if (event->event_code != HCI_EVENT_CODE_COMMAND_COMPLETE_EVENT) {
    hci_bs_printf("HCI_EVENT_CODE_COMMAND_COMPLETE_EVENT failed\n");
    return -1;
  }
  event->command_length = *buf++;
  if (event->command_length != 0x04) {
    hci_bs_printf("expected length != 4\n");
    return -1;
  }
  event->num_hci_packets = *buf++;
  event->ocf_ogf = ((buf[0] << 8) | buf[1]);
  buf += 2;
  event->status = *buf++;

  return 0;
}

static int parse_hci_command_event(uint8_t *buf, int len, struct hci_command_event *event)
{
  event->packet_type = *buf++;
  if (event->packet_type != HCI_PACKET_TYPE_EVENT_PACKET) {
    hci_bs_printf("HCI_PACKET_TYPE_EVENT_PACKET failed\n");
    return -1;
  }
  event->event_code = *buf++;
  event->command_length = *buf++;
  event->reserved1 = *buf++;
  event->reserved2 = *buf++;
  event->command = *buf++;
  event->status = *buf++;

  return 0;
}

int parse_hci_m2_xet_event(uint8_t *buf, int len, struct hci_m2_xet_event *event)
{
  event->packet_type = *buf++;
  if (event->packet_type != HCI_PACKET_TYPE_EVENT_PACKET) {
    hci_bs_printf("HCI_PACKET_TYPE_EVENT_PACKET failed\n");
    return -1;
  }
  event->event_code = *buf++;
  if (event->event_code != 0xff) {
    hci_bs_printf("0xff failed\n");
    return -1;
  }
  event->command_length = *buf++;

  event->reserved1 = *buf++;
  event->reserved2 = *buf++;
  event->command = *buf++;
  event->reserved[0] = *buf++;
  event->reserved[1] = *buf++;
  event->reserved[2] = *buf++;
  event->acceptor = *buf++;
  event->information_id = *buf++;
  event->command_result = *buf++;
  event->data_type = *buf++;
  event->data = buf;

  return 0;
}

static int wait_for_command_complete_event(uint16_t expected_ocf_ogf)
{
  int len;
  uint8_t *ptr = uart_buffer;
  struct hci_command_complete_event event;
  int ret;

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_command_complete_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_command_complete_event() failed\n");
    return -1;
  }

  if (event.ocf_ogf != expected_ocf_ogf) {
    hci_bs_printf("expected ocf 0x%02x, got 0x%0x2\n", expected_ocf_ogf, event.ocf_ogf);
    return -1;
  }

  if (event.status) {
    hci_bs_printf("status 0x%02x is not 0\n", ptr[6]);
    return -1;
  }

  hci_bs_printf("expected_ocf_ogf 0x%04x\n", expected_ocf_ogf);
  return 0;
}

int hci_api_reset(void)
{
  const uint8_t request[] = {0x01, SET_OCF_OGF(OCF_OGF_HCI_RESET), 0x00};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = wait_for_command_complete_event(OCF_OGF_HCI_RESET);
  if (ret) {
    hci_bs_printf("wait_for_command_complete_event() failed\n");
    return ret;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_hardware_reset_handler(void)
{
  int ret;

  ret = wait_for_command_complete_event(0x0000);
  if (ret) {
    hci_bs_printf("wait_for_command_complete_event() failed\n");
    return ret;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_write_bd_addr(uint8_t *bd_addr)
{
  uint8_t request[] = {0x01, SET_OCF_OGF(OCF_OGF_HCI_WRITE_BD_ADDRESS), 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  int ret;

  memcpy(&request[4], bd_addr, 6);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = wait_for_command_complete_event(OCF_OGF_HCI_WRITE_BD_ADDRESS);
  if (ret) {
    hci_bs_printf("wait_for_command_complete_event() failed\n");
    return ret;
  }

  hci_bs_printf("\n");
  return 0;
}

static const uint8_t TC35661_501_ROM_STR[] = "8.00.72B-06 ROM=501";
static const uint8_t TC35661_501_ROM_LENGTH = sizeof(TC35661_501_ROM_STR);
static const uint8_t TC35661_502_ROM_STR[] = "8.00.72D-12 ROM=502";
static const uint8_t TC35661_502_ROM_LENGTH = sizeof(TC35661_502_ROM_STR);
static const uint8_t TC35667_004_ROM_STR[] = "C.00.85D-02 ROM=004";
static const uint8_t TC35667_004_ROM_LENGTH = sizeof(TC35667_004_ROM_STR);
static const uint8_t TC35667_005_ROM_1_STR[] = "C.00.85D-02 ROM=005";
static const uint8_t TC35667_005_ROM_LENGTH_1 = sizeof(TC35667_005_ROM_1_STR);
static const uint8_t TC35667_005_ROM_2_STR[] = "C.00.85E-19 ROM=005";
static const uint8_t TC35667_005_ROM_LENGTH_2 = sizeof(TC35667_005_ROM_2_STR);
static const uint8_t TC35667_006_ROM_STR[] = "D.00.85A-12 ROM=006";
static const uint8_t TC35667_006_ROM_LENGTH = sizeof(TC35667_006_ROM_STR);
static const uint8_t TC35670_003_ROM_STR[] = "C.00.85C-04 ROM=003";
static const uint8_t TC35670_003_ROM_LENGTH = sizeof(TC35670_003_ROM_STR);
static const uint8_t TC35675_001_ROM_STR[] = "D.00.86B-05 ROM=001";
static const uint8_t TC35675_001_ROM_LENGTH = sizeof(TC35675_001_ROM_STR);
static const uint8_t TC35677_8_9_002_ROM_STR[] = "F.00.87C-09 ROM=002";
static const uint8_t TC35677_8_9_002_ROM_LENGTH = sizeof(TC35677_8_9_002_ROM_STR);
static const uint8_t TC35661_551_ROM_STR[] = "F.02.72H-07 ROM=551";
static const uint8_t TC35661_551_ROM_LENGTH = sizeof(TC35661_551_ROM_STR);
static const uint8_t TC35661_503_ROM_STR[] = "F.01.72F-02 ROM=503";
static const uint8_t TC35661_503_ROM_LENGTH = sizeof(TC35661_503_ROM_STR);
static const uint8_t TC3567C_002_ROM_STR[] = "F.00.89B-06 ROM=002";
static const uint8_t TC3567C_002_ROM_LENGTH = sizeof(TC3567C_002_ROM_STR);
static const uint8_t TC35680_1_002_ROM_STR[] = "G.00.B0B-03 ROM=002";
static const uint8_t TC35680_1_002_ROM_LENGTH = sizeof(TC35680_1_002_ROM_STR);

int hci_api_read_firmware_version(enum tsb_bluetooth_chip_version *version)
{
  const uint8_t request[] = {0x01, 0x08, 0xfc, 0x09, 0x00, M2_GET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_GET_FIRMWARE_VERSION, 0xff, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  if (version == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_GET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != CMD_GET_FIRMWARE_VERSION) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  if (event.data_type != 0x0f) {
    hci_bs_printf("unexpected data_type 0x%02x\n", event.data_type);
    return -1;
  }

  if (!memcmp(&event.data[0], TC35661_501_ROM_STR, TC35661_501_ROM_LENGTH)) {
    *version = TC35661_501_ROM;
  } else if (!memcmp(&event.data[0], TC35661_502_ROM_STR, TC35661_502_ROM_LENGTH)) {
    *version = TC35661_502_ROM;
  } else if (!memcmp(&event.data[0], TC35667_004_ROM_STR, TC35667_004_ROM_LENGTH)) {
    *version = TC35667_004_ROM;
  } else if (!memcmp(&event.data[0], TC35667_005_ROM_1_STR, TC35667_005_ROM_LENGTH_1)) {
    *version = TC35667_005_ROM_1;
  } else if (!memcmp(&event.data[0], TC35667_005_ROM_2_STR, TC35667_005_ROM_LENGTH_2)) {
    *version = TC35667_005_ROM_2;
  } else if (!memcmp(&event.data[0], TC35667_006_ROM_STR, TC35667_006_ROM_LENGTH)) {
    *version = TC35667_006_ROM;
  } else if (!memcmp(&event.data[0], TC35670_003_ROM_STR, TC35670_003_ROM_LENGTH)) {
    *version = TC35670_003_ROM;
  } else if (!memcmp(&event.data[0], TC35675_001_ROM_STR, TC35675_001_ROM_LENGTH)) {
    *version = TC35675_001_ROM;
  } else if (!memcmp(&event.data[0], TC35677_8_9_002_ROM_STR, TC35677_8_9_002_ROM_LENGTH)) {
    *version = TC35677_8_9_002_ROM;
  } else if (!memcmp(&event.data[0], TC35661_551_ROM_STR, TC35661_551_ROM_LENGTH)) {
    *version = TC35661_551_ROM;
  } else if (!memcmp(&event.data[0], TC35661_503_ROM_STR, TC35661_503_ROM_LENGTH)) {
    *version = TC35661_503_ROM;
  } else if (!memcmp(&event.data[0], TC3567C_002_ROM_STR, TC3567C_002_ROM_LENGTH)) {
    *version = TC3567C_002_ROM;
  } else if (!memcmp(&event.data[0], TC35680_1_002_ROM_STR, TC35680_1_002_ROM_LENGTH)) {
    *version = TC35680_1_002_ROM;
  } else {
    bs_warning("unknown version '%s'\n", &event.data[0]);
    *version = UNKNOWN_VERSION;
  }

  bs_warning("version '%s'\n", &event.data[0]);

  return 0;
}

int hci_api_set_deep_sleep(const struct hci_api_deep_sleep_data *data)
{
  uint8_t request[] = {
      0x01, 0x08, 0xfc, 0x1c, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_SET_DEEP_SLEEP, 0xff, 0x10, 0x12, 0x00, 0xb8, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
  int ret;

  if (data == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[14] = data->clkreq_signal;
  request[19] = (uint8_t)(1 << data->notify_interface);
  request[20] = (uint8_t)(data->osciallator_drift & 0xff);
  request[21] = (uint8_t)((data->osciallator_drift >> 8) & 0xff);
  request[22] = (uint8_t)(data->jitter_drift & 0xff);
  request[23] = (uint8_t)((data->jitter_drift >> 8) & 0xff);
  request[24] = (uint8_t)(data->sniff_attempt_wakeup_margin & 0xff);
  request[25] = (uint8_t)((data->sniff_attempt_wakeup_margin >> 8) & 0xff);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_SET_DEEP_SLEEP);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_uart_rts_cts_control(int enable)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x03, 0x00, HCI_RTS_CTS, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_command_event event;
  int len;
  int ret;

  /* make sure that enable byte is binary */
  request[6] = (uint8_t)(!!enable);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_command_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_command_event() failed\n");
    return -1;
  }

  if (event.command != HCI_RTS_CTS) {
    hci_bs_printf("expected HCI_RTS_CTS, got 0x%0x2\n", event.command);
    return -1;
  }

  if (event.status != 0) {
    hci_bs_printf("status %d\n", event.status);
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_switch_to_tcu_mode(enum tsb_bluetooth_chip_version version)
{
  const uint8_t request[] = {0x01, 0x08, 0xfc, 0x03, 0x00, HCI_SET_MODE, 0x01};
  uint8_t *ptr = uart_buffer;
  struct hci_command_event event;
  int len;
  int ret;

  if (version == TC35661_501_ROM) {
    ret = hci_api_tc35661_update_firmware();
    if (ret) {
      hci_bs_printf("hci_api_tc35661_update_firmware() failed\n");
      return -1;
    }
  }

  if (version == TC35667_005_ROM_2) {
    ret = hci_api_tc35667_005_update_firmware();
    if (ret) {
      hci_bs_printf("hci_api_tc35667_005_update_firmware() failed\n");
      return -1;
    }
  }

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_command_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_command_event() failed\n");
    return -1;
  }

  if (event.command != HCI_SET_MODE) {
    hci_bs_printf("expected HCI_SET_MODE, got 0x%0x2\n", event.command);
    return -1;
  }

  if (event.status != 0) {
    hci_bs_printf("status %d\n", event.status);
    return -1;
  }

  hci_mode_active = 0;

  hci_bs_printf("\n");
  return 0;
}

static int check_m2(uint8_t cmd)
{
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_SET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != cmd) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_set_i2_enable(void)
{
  const uint8_t request[] = {0x01, 0x08, 0xfc, 0x0b, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_I2C_ENABLE, 0xff, 0x02, 0x01, 0x01};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_I2C_ENABLE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_eeprom_write_enable_tc35661(void)
{
  const uint8_t request[] = {0x01, 0x08, 0xfc, 0x09, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_EEPROM_WRITE_ENABLE, 0xff, 0x00};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_EEPROM_WRITE_ENABLE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_write_eeprom(uint8_t eeprom_device_address, uint8_t addressing_16bit, uint16_t offset, uint8_t *buf, uint8_t write_length)
{
#define MAX_WRITE_LENGTH 32
  uint8_t request[20 + MAX_WRITE_LENGTH] = {0x01, 0x08, 0xfc, 0x10, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_EEPROM_WRITE, 0xff, 0x10, 0x06, 0xa0, 0x01, 0x01, 0x06, 0x02, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  if ((write_length > MAX_WRITE_LENGTH) || (write_length == 0)) {
    hci_bs_printf("write_length %d out of bounds\n", write_length);
    return -1;
  }

  if (buf == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[3] = 0x10 + write_length;
  request[13] = 0x06 + write_length;
  request[14] = eeprom_device_address;
  request[15] = !!addressing_16bit;
  request[17] = write_length;
  request[18] = (uint8_t)(offset & 0xff);
  request[19] = (uint8_t)((offset >> 8) & 0xff);
  memcpy(&request[20], buf, write_length);

  ret = uart_application_send_bytes(&request[0], 20 + write_length);
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_SET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != CMD_EEPROM_WRITE) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  if (event.data_type != 0x00) {
    hci_bs_printf("unexpected data_type 0x%02x\n", event.data_type);
    return -1;
  }

  return 0;
}

int hci_api_read_eeprom(uint8_t eeprom_device_address, uint8_t addressing_16bit, uint16_t offset, uint8_t *buf, uint8_t read_length)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x10, 0x00, M2_GET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_EEPROM_READ, 0xff, 0x10, 0x06, 0xa0, 0x01, 0x01, 0x06, 0x02, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  if (read_length == 0) {
    hci_bs_printf("read_length %d out of bounds\n", read_length);
    return -1;
  }

  if (buf == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[14] = eeprom_device_address;
  request[15] = !!addressing_16bit;
  request[17] = read_length;
  request[18] = (uint8_t)(offset & 0xff);
  request[19] = (uint8_t)((offset >> 8) & 0xff);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_GET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != CMD_EEPROM_READ) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  if (event.data_type != 0x10) {
    hci_bs_printf("unexpected data_type 0x%02x\n", event.data_type);
    return -1;
  }

  if (event.data[0] != read_length) {
    hci_bs_printf("unexpected read length %d\n", event.data[0]);
    return -1;
  }

  memcpy(buf, &event.data[1], read_length);

  return 0;
}

int hci_api_execute_dloc_write(const struct hci_api_dbus_write_command *command)
{
  uint8_t request[] = {
      0x01, 0x03, 0xfc, 0x05, 0x00, HCI_LOC_DBUS_WRITE_CMD, 0x00, 0x00, 0x00,
  };
  uint8_t *ptr = uart_buffer;
  struct hci_command_event event;
  int len;
  int ret;

  if (command == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[6] = command->address;
  request[7] = command->lsb;
  request[8] = command->msb;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  if (ptr[1] != 0x0f) {
    hci_bs_printf("unexpected event_code 0x%02x\n", ptr[1]);
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_command_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != HCI_LOC_DBUS_WRITE_CMD) {
    hci_bs_printf("expected HCI_SET_MODE, got 0x%0x2\n", event.command);
    return -1;
  }

  if (event.status != 0) {
    hci_bs_printf("status %d\n", event.status);
    return -1;
  }

  hci_bs_printf("address 0x%08x\n", command->address);

  return 0;
}

int hci_api_execute_m2_set(const uint8_t *request, int length)
{
  int ret;

  ret = uart_application_send_bytes(&request[0], length);
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(request[10]);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

#define CMD_BTL_SET_BAUDRATE 0x42

static int set_baudrate_tc35661(enum hci_api_baudrate baudrate)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x11, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_BTL_SET_BAUDRATE, 0xff, 0x10, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  uint16_t div1;
  uint8_t div2;
  int ret;

  switch (baudrate) {
  case HCI_API_BAUDRATE_115200:
    div1 = 0x001A;
    div2 = 0x60;
    break;
  case HCI_API_BAUDRATE_230400:
    div1 = 0x000D;
    div2 = 0x60;
    break;
  case HCI_API_BAUDRATE_460800:
    div1 = 0x0005;
    div2 = 0xA0;
    break;
  case HCI_API_BAUDRATE_921600:
    div1 = 0x0003;
    div2 = 0x70;
    break;
  default:
    return -1;
  }

  request[14] = (uint8_t)(0xff & div1);
  request[15] = (uint8_t)((0xff00 & div1) >> 8);
  request[16] = div2;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_api_set_baudrate() failed\n");
    return -1;
  }

  ret = check_m2(CMD_BTL_SET_BAUDRATE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  return 0;
}

#define DIVISION_RATIO_12 0x50
#define DIVISION_RATIO_13 0x60
#define DIVISION_RATIO_14 0x70
#define DIVISION_RATIO_15 0x80
#define DIVISION_RATIO_16 0x90
#define DIVISION_RATIO_17 0xa0

static int set_baudrate_tc35667(enum hci_api_baudrate baudrate)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x11, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_BTL_SET_BAUDRATE, 0xff, 0x10, 0x07, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00};
  uint16_t div1;
  uint8_t div2;
  int ret;

  switch (baudrate) {
  case HCI_API_BAUDRATE_115200:
    div1 = 15;
    div2 = DIVISION_RATIO_15;
    break;
  case HCI_API_BAUDRATE_230400:
    div1 = 7;
    div2 = DIVISION_RATIO_14;
    break;
  case HCI_API_BAUDRATE_460800:
    div1 = 4;
    div2 = DIVISION_RATIO_14;
    break;
  case HCI_API_BAUDRATE_921600:
    div1 = 2;
    div2 = DIVISION_RATIO_14;
    break;
  default:
    return -1;
  }

  request[14] = (uint8_t)(0xff & div1);
  request[15] = (uint8_t)((0xff00 & div1) >> 8);
  request[16] = div2;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_api_set_baudrate() failed\n");
    return -1;
  }

  ret = check_m2(CMD_BTL_SET_BAUDRATE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  return 0;
}

static int set_baudrate_tc35680(enum hci_api_baudrate baudrate)
{
  uint8_t request[] = { 0x01, 0x08, 0xfc, 0x11, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_BTL_SET_BAUDRATE, 0xff, 0x10, 0x07, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00 };
  int ret;

  switch (baudrate) {
  case HCI_API_BAUDRATE_115200:
    request[16] = 0x50;
    request[14] = 0x20;
    request[15] = 0x00;
    request[18] = 0x00;
    break;
  case HCI_API_BAUDRATE_230400:
    request[16] = 0x50;
    request[14] = 0x0b;
    request[15] = 0x00;
    request[18] = 0x25;
    break;
  case HCI_API_BAUDRATE_460800:
    request[16] = 0x60;
    request[14] = 0x05;
    request[15] = 0x00;
    request[18] = 0x16;
    break;
  case HCI_API_BAUDRATE_921600:
    request[16] = 0x90;
    request[14] = 0x02;
    request[15] = 0x00;
    request[18] = 0x0b;
    break;
  default:
    return -1;
  }

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_api_set_baudrate() failed\n");
    return -1;
  }

  ret = check_m2(CMD_BTL_SET_BAUDRATE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  return 0;
}

int hci_api_set_baudrate(enum tsb_bluetooth_chip_version version, enum hci_api_baudrate baudrate)
{
  if (version == TC35661_501_ROM || version == TC35661_502_ROM) {
    return set_baudrate_tc35661(baudrate);
  }
  if (version == TC35680_1_001_ROM || version == TC35680_1_002_ROM) {
    return set_baudrate_tc35680(baudrate);
  }
  return set_baudrate_tc35667(baudrate);
}
 

int hci_api_set_rtscts(void)
{
  const uint8_t request[] = {0x01, 0x08, 0xfc, 0x0a, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_BLE_SET_RTSCTS, 0xff, 0x01, 0xff};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_BLE_SET_RTSCTS);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_set_tx_power_tc35667(enum hci_api_transmit_power power)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x0b, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_BLE_SET_TRANSMIT_POWER, 0xff, 0x10, 0x01, 0x00};
  int ret;

  if (power < HCI_API_TRANSMIT_POWER_0DBM || power > HCI_API_TRANSMIT_POWER_MINUS20DBM) {
    hci_bs_printf("power %d out of bounds\n", power);
    return -1;
  }

  request[14] = power;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_BLE_SET_TRANSMIT_POWER);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_open_storage_flash(int write_protection)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x0f, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_FLASH_OPEN_STORAGE, 0xff, 0x10, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00};
  int ret;

  request[15] = (uint8_t)((write_protection >> 24) & 0xff);
  request[16] = (uint8_t)((write_protection >> 16) & 0xff);
  request[17] = (uint8_t)((write_protection >> 8) & 0xff);
  request[18] = (uint8_t)((write_protection) & 0xff);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_FLASH_OPEN_STORAGE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_read_data_from_storage_device(uint32_t address, uint16_t length, uint8_t *buf)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x10, 0x00, M2_GET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_FLASH_READ_WRITE_STORAGE, 0xff, 0x10, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  if (length == 0) {
    hci_bs_printf("length %d out of bounds\n", length);
    return -1;
  }

  if (buf == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[14] = (uint8_t)(0x000000ff & address);
  request[15] = (uint8_t)((0x0000ff00 & address) >> 8);
  request[16] = (uint8_t)((0x00ff0000 & address) >> 16);
  request[17] = (uint8_t)((0xff000000 & address) >> 24);
  request[18] = (uint8_t)(0xff & length);
  request[19] = (uint8_t)((0xff00 & length) >> 8);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_GET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != CMD_FLASH_READ_WRITE_STORAGE) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  if (event.data_type != 0x10) {
    hci_bs_printf("unexpected data_type 0x%02x\n", event.data_type);
    return -1;
  }

  if (event.data[0] != length) {
    hci_bs_printf("unexpected read length %d\n", event.data[0]);
    return -1;
  }

  memcpy(buf, &event.data[1], length);

  return 0;
}

int hci_api_write_data_to_storage_device(uint32_t address, uint16_t length, uint8_t *buf)
{
#define MAX_LENGTH 208
  uint8_t request[20 + MAX_LENGTH] = {0x01, 0x08, 0xfc, 0x10, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_FLASH_READ_WRITE_STORAGE, 0xff, 0x10, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  uint8_t *ptr = uart_buffer;
  struct hci_m2_xet_event event;
  int len;
  int ret;

  if ((length > MAX_LENGTH) || (length == 0)) {
    hci_bs_printf("length %d out of bounds\n", length);
    return -1;
  }

  if (buf == NULL) {
    hci_bs_printf("NULL Pointer Error\n");
    return -1;
  }

  request[3] = ((uint8_t)0x10) + ((uint8_t)length);
  request[13] = ((uint8_t)0x06) + ((uint8_t)length);
  request[14] = (uint8_t)(0x000000ff & address);
  request[15] = (uint8_t)((0x0000ff00 & address) >> 8);
  request[16] = (uint8_t)((0x00ff0000 & address) >> 16);
  request[17] = (uint8_t)((0xff000000 & address) >> 24);
  request[18] = (uint8_t)(0xff & length);
  request[19] = (uint8_t)((0xff00 & length) >> 8);

  memcpy(&request[20], buf, length);

  ret = uart_application_send_bytes(&request[0], 20 + length);
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  len = hci_receive_response(ptr, uart_buffer_len);
  if (len < 0) {
    hci_bs_printf("hci_receive_response() failed\n");
    return -1;
  }

  ret = parse_hci_m2_xet_event(ptr, len, &event);
  if (ret) {
    hci_bs_printf("parse_hci_m2_xet_event() failed\n");
    return -1;
  }

  if (event.command != M2_SET_CMD) {
    hci_bs_printf("unexpected command 0x%02x\n", event.command);
    return -1;
  }

  if (event.information_id != CMD_FLASH_READ_WRITE_STORAGE) {
    hci_bs_printf("unexpected information_id 0x%02x\n", event.information_id);
    return -1;
  }

  if (event.command_result != 0) {
    hci_bs_printf("unexpected command_result 0x%02x\n", event.command_result);
    return -1;
  }

  if (event.data_type != 0x00) {
    hci_bs_printf("unexpected data_type 0x%02x\n", event.data_type);
    return -1;
  }

  return 0;
}

int hci_api_erase_sector_in_storage_device(uint32_t address)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x0e, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_FLASH_ERASE_SECTOR_IN_STORAGE, 0xff, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00};
  int ret;

  request[14] = (uint8_t)(0x000000ff & address);
  request[15] = (uint8_t)((0x0000ff00 & address) >> 8);
  request[16] = (uint8_t)((0x00ff0000 & address) >> 16);
  request[17] = (uint8_t)((0xff000000 & address) >> 24);

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_FLASH_ERASE_SECTOR_IN_STORAGE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_erase_storage_device(void)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x09, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x94, CMD_FLASH_ERASE_STORAGE_DEVICE, 0xff, 0x00};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_FLASH_ERASE_STORAGE_DEVICE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_api_close_storage_flash(void)
{
  uint8_t request[] = {0x01, 0x08, 0xfc, 0x09, 0x00, M2_SET_CMD, 0x00, 0x00, 0x00, 0x14, CMD_FLASH_CLOSE_STORAGE, 0xff, 0x00};
  int ret;

  ret = uart_application_send_bytes(&request[0], sizeof(request));
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = check_m2(CMD_FLASH_CLOSE_STORAGE);
  if (ret) {
    hci_bs_printf("check_m2() failed\n");
    return -1;
  }

  hci_bs_printf("\n");
  return 0;
}

int hci_receive_response(uint8_t *buf, uint32_t len)
{
  uint8_t command_length;
  uint8_t msg_len;
  int ret;

  if (len < 3) {
    hci_bs_printf("hci_receive_response() @ hci response received with invalid length\n");
    return -1;
  }

  ret = uart_application_receive_receive_bytes(buf, 3, 1000);
  if (ret) {
    hci_bs_printf("uart_application_receive_receive_bytes() failed\n");
    return -1;
  }

  command_length = buf[2];

  if (len < (uint8_t)(command_length + 3)) {
    hci_bs_printf("len %d out of bounds\n", len);
    return -1;
  }

  ret = uart_application_receive_receive_bytes(buf + 3, command_length, 100);
  if (ret) {
    hci_bs_printf("uart_application_receive_receive_bytes() failed\n");
    return -1;
  }

  data_receive_debug_callback(buf, (uint32_t)(command_length + 3));

  ret = HCI_API_CHECK_HCI_RESPONSE(buf);
  if (ret) {
    hci_bs_printf("HCI_API_CHECK_HCI_RESPONSE failed\n");
    return -1;
  }

  msg_len = HCI_API_RETRIEVE_HCI_RESPONSE_LENGTH(buf);
  if (msg_len > len) {
    hci_bs_printf("msg_len > len\n");
    return -1;
  }

  return msg_len;
}

int hci_api_set_extended_inquiry_response(const uint8_t *name, uint16_t len)
{
#define MAX_DEV_NAME_LEN_EIR 238 /* Extended Inquiry Response (240 octets) - Length - EIR Data Type = 240 - 1 - 1 = 238 */
  uint8_t *ptr = uart_buffer;
  int ret;

  if (uart_buffer_len < 245) {
    hci_bs_printf("uart_buffer_len %d too small\n", uart_buffer_len);
    return -1;
  }

  if (len > MAX_DEV_NAME_LEN_EIR) {
    hci_bs_printf("len %d out of range\n", len);
    return -1;
  }

  memset(ptr, 0, uart_buffer_len);

  *ptr++ = (uint8_t)0x01; /* hci command packet */
  *ptr++ = (uint8_t)0x52; /* ocf = write extended inquiry response */
  *ptr++ = (uint8_t)0x0c; /* ocf cont. - ogf = controller and baseband command */
  *ptr++ = (uint8_t)241;  /* total length */
  *ptr++ = (uint8_t)0x00; /* fec is not required */

  /* now 240 bytes eir to follow */
  *ptr++ = (uint8_t)(len + 1);  /* length next ad entry*/
  *ptr++ = (uint8_t)0x09;       /* Complete local name */

  memcpy(ptr, name, len);
  ptr += len;

  ret = uart_application_send_bytes(uart_buffer, 245);
  if (ret) {
    hci_bs_printf("hci_send_request() failed\n");
    return -1;
  }

  ret = wait_for_command_complete_event(0x520c);
  if (ret) {
    hci_bs_printf("wait_for_command_complete_event() failed\n");
    return ret;
  }

  return 0;
}

