/*
 * COPYRIGHT (C) 2018
 * TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION ALL RIGHTS RESERVED
 *
 * THE SOURCE CODE AND ITS RELATED DOCUMENTATION IS PROVIDED "AS IS". TOSHIBA
 * ELECTRONIC DEVICES & STORAGE CORPORATION MAKES NO OTHER WARRANTY OF ANY KIND,
 * WHETHER EXPRESS, IMPLIED OR, STATUTORY AND DISCLAIMS ANY AND ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY, SATISFACTORY QUALITY, NON INFRINGEMENT AND
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * THE SOURCE CODE AND DOCUMENTATION MAY INCLUDE ERRORS. TOSHIBA ELECTRONIC
 * DEVICES & STORAGE CORPORATION RESERVES THE RIGHT TO INCORPORATE MODIFICATIONS
 * TO THE SOURCE CODE IN LATER REVISIONS OF IT, AND TO MAKE IMPROVEMENTS OR
 * CHANGES IN THE DOCUMENTATION OR THE PRODUCTS OR TECHNOLOGIES DESCRIBED
 * THEREIN AT ANY TIME.
 *
 * TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION SHALL NOT BE LIABLE FOR ANY
 * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGE OR LIABILITY ARISING FROM YOUR USE
 * OF THE SOURCE CODE OR ANY DOCUMENTATION, INCLUDING BUT NOT LIMITED TO, LOST
 * REVENUES, DATA OR PROFITS, DAMAGES OF ANY SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL NATURE, PUNITIVE DAMAGES, LOSS OF PROPERTY OR LOSS OF PROFITS
 * ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT, OR BEING UNUSABLE, EVEN
 * IF ADVISED OF THE POSSIBILITY OR PROBABILITY OF SUCH DAMAGES AND WHETHER
 * A CLAIM FOR SUCH DAMAGE IS BASED UPON WARRANTY, CONTRACT, TORT, NEGLIGENCE OR
 * OTHERWISE.
 */

/**
* @file hci_api.h
*
* @brief Contains definitions and function declarations for Bluetooth HCI API.
*/
#ifndef __HCI_API_H__
#define __HCI_API_H__

#include "uart_api.h"
#include "types.h"
#include "versions.h"

/*! Signals if HCI mode is active or not */
extern int hci_mode_active;

/*! HCI deep sleep CLKREQ signal enumeration */
enum hci_api_deep_sleep_clkreq_signal {
  hci_clkreq_work_deep_sleep = 0x00, /*!< work deep sleep */
  hci_clkreq_always_low = 0x01,      /*!< always low */
  hci_clkreq_always_high = 0x02      /*!< always high */
};

/*! HCI deep sleep CLKREQ signal enumeration */
enum hci_api_deep_sleep_notify_interface {
  hci_notify_interface_gpio = 0, /*!< GPIO */
  hci_notify_interface_uart = 1, /*!< UART */
  hci_notify_interface_usb = 2   /*!< USB */
};

/*! HCI UART baudrate enumeration */
enum hci_api_baudrate {
  HCI_API_BAUDRATE_115200, /*!< 115200 */
  HCI_API_BAUDRATE_230400, /*!< 230400 */
  HCI_API_BAUDRATE_460800, /*!< 460800 */
  HCI_API_BAUDRATE_921600  /*!< 921600 */
};

/*! HCI UART transmit power enumeration */
enum hci_api_transmit_power {
  HCI_API_TRANSMIT_POWER_0DBM = 0x01,       /*!< 0dBm output power */
  HCI_API_TRANSMIT_POWER_MINUS4DBM = 0x02,  /*!< -4dBm output power */
  HCI_API_TRANSMIT_POWER_MINUS8DBM = 0x03,  /*!< -8dBm output power */
  HCI_API_TRANSMIT_POWER_MINUS12DBM = 0x04, /*!< -12dBm output power */
  HCI_API_TRANSMIT_POWER_MINUS16DBM = 0x05, /*!< -16dBm output power */
  HCI_API_TRANSMIT_POWER_MINUS20DBM = 0x06  /*!< -20dBm output power */
};

/*! Structure definition for HCI deep sleep configuration data */
struct hci_api_deep_sleep_data
{
  enum hci_api_deep_sleep_clkreq_signal clkreq_signal;       /*!< clkreq signal */
  enum hci_api_deep_sleep_notify_interface notify_interface; /*!< notify interface */
  uint16_t osciallator_drift;                                /*!< osciallator drift */
  uint16_t jitter_drift;                                     /*!< jitter drift */
  uint16_t sniff_attempt_wakeup_margin;                      /*!< sniff attempt wakeup margin*/
};

/**
* @brief Reset the Bluetooth HCI controller.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_reset(void);

/**
* @brief Enable RTS/CTS handling.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_set_rtscts(void);

/**
* @brief Set Bluetooth device address in HCI controller.
*
* @param[in] *bd_addr Pointer to Bluetooth device address.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_write_bd_addr(uint8_t *bd_addr);

/**
* @brief Switch the Bluetooth controller from HCI to Toshiba TCU mode.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_switch_to_tcu_mode(enum tsb_bluetooth_chip_version version);

/**
* @brief Enable or disable RTS/CTS control in TC3566x
*
* @param[in] enable Set to enable, otherwise disable
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_uart_rts_cts_control(int enable);

/**
* @brief Enable I2C subsystem on HCI controller.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_set_i2_enable(void);

/**
* @brief Enables writing in I2C subsystem on HCI controller. Only necessary
* for TC35661.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_eeprom_write_enable_tc35661(void);

/**
* @brief Reads data from the EEPROM attached to the HCI controller.
*
* @param[in] eeprom_device_address Device address of eeprom
* @param[in] addressing_16bit Indicates if 16bit addressing must be used
* @param[in] offset Read offset into the EEPROM
* @param[in] *buf Pointer to buffer to receive read data
* @param[in] read_length Number of bytes to read
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_read_eeprom(uint8_t eeprom_device_address, uint8_t addressing_16bit, uint16_t offset, uint8_t *buf, uint8_t read_length);

/**
* @brief Writes data to the EEPROM attached to the HCI controller.
*
* The maximum number of bytes that can be written at once is 32 bytes.
*
* @param[in] eeprom_device_address Device address of eeprom
* @param[in] addressing_16bit Indicates if 16bit addressing must be used
* @param[in] offset Write offset into the EEPROM
* @param[in] *buf Pointer to buffer to data to be written
* @param[in] write_length Number of bytes to write
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_write_eeprom(uint8_t eeprom_device_address, uint8_t addressing_16bit, uint16_t offset, uint8_t *buf, uint8_t write_length);

/**
* @brief Read out the firmware version string (zero terminated) from
* the HCI controller.
*
* The pointer is valid until the next call to the HCI API.
*
* @param[out] *version Pointer that will point to the firmware version identifier
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_read_firmware_version(enum tsb_bluetooth_chip_version *version);

/**
* @brief Send HCI deep sleep configuration command to the HCI controller.
*
* @param[in] *data Pointer to deep sleep configuration data
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_set_deep_sleep(const struct hci_api_deep_sleep_data *data);

/**
* @brief Clean up after a hardware reset of the HCI controller.
*
* When the HCI controller is reset by a hardware signal, it will
* output a "reset done" message. This function will check this
* message and process it correctly.
*
* @return 0 Success
* @return -1 Error
*/
int hci_api_hardware_reset_handler(void);

/**
* @brief Configure a different baudrate for TC35667
*
* @param[in] baudrate New baudrate to be used
*/
int hci_api_set_baudrate(enum tsb_bluetooth_chip_version version, enum hci_api_baudrate baudrate);

/**
* @brief Configure a different tx power rate for TC35667
*
* @param[in] power New tx power to be used
*/
int hci_api_set_tx_power_tc35667(enum hci_api_transmit_power power);

/**
* @brief command to open Non-volatile memory in the TC35675 and TC35678
*
*        Below values can be provided as write_protect
*        0x00000000: No protect
*        0x64000000: 0x000000-0x000fff 4KB 1/64
*        0x68000000: 0x000000-0x001fff 8KB 2/64
*        0x6C000000: 0x000000-0x003fff 16KB 4/64
*        0x70000000: 0x000000-0x007fff 32KB 8/64
*        0x24000000: 0x000000-0x00ffff 64KB 16/64
*        0x28000000: 0x000000-0x01ffff 128KB 32/64
*        0x04400000: 0x000000-0x02ffff 192KB 48/64
*        0x50400000: 0x000000-0x037fff 224KB 56/64
*        0x4C400000: 0x000000-0x03bfff 240KB 60/64
*        0x48400000: 0x000000-0x03dfff 248KB 62/64
*        0x44400000: 0x000000-0x03efff 252KB 63/64
*        0x0C000000: 0x000000-0x03ffff 256KB 64/64 (Read access only)
*
* @param[in] write_protection Write protect setting
*
*/
int hci_api_open_storage_flash(int write_protection);

/**
* @brief command to read Non-volatile memory in the TC35675 and TC35678
*
* @param[in] address  Memory address from where data to be read
* @param[in] length   Number of bytes to be read
* @param[out] *buf    Pointer to buffer to receive read data
*/
int hci_api_read_data_from_storage_device(uint32_t address, uint16_t length, uint8_t *buf);

/**
* @brief command to write to Non-volatile memory in the TC35675 and TC35678
*
* @param[in] address  Memory address where data to be written
* @param[in] length   Number of bytes to be written
* @param[out] *buf    Pointer to buffer to data to be written
*/
int hci_api_write_data_to_storage_device(uint32_t address, uint16_t length, uint8_t *buf);

/**
* @brief command to erase all data in Non-volatile memory in the TC35675 and TC35678
*
*/
int hci_api_erase_storage_device(void);

/**
* @brief command to Erase sector in Non-volatile memory in the TC35675 and TC35678
*
* @param[in] address  Starting address of sector to be erased
*/
int hci_api_erase_sector_in_storage_device(uint32_t address);

/**
* @brief command to Close Non-volatile memory in the TC35675 and TC35678
*
*/
int hci_api_close_storage_flash(void);

/**
* @brief command to set the EIR
*
* @param[in] *name Pointer to Complete local name
* @param[in] len Length of Complete local name
*/
int hci_api_set_extended_inquiry_response(const uint8_t *name, uint16_t len);

#endif

